@@ -3,75 +3,57 @@ __copyright__ = "(c) Oleg Butovich 2013-2017"
 __license__ = "MIT"
 
 
+import io
 import json
 import logging
 import os
+import platform
 import sys
 import time
 from shlex import split as shell_split
 
-from proxmoxer.core import SERVICES, config_failure
+from proxmoxer.core import SERVICES, AuthenticationError, config_failure
 
 logger = logging.getLogger(__name__)
 logger.setLevel(level=logging.WARNING)
 
-STREAMING_SIZE_THRESHOLD = 100 * 1024 * 1024  # 10 MiB
+STREAMING_SIZE_THRESHOLD = 10 * 1024 * 1024  # 10 MiB
+SSL_OVERFLOW_THRESHOLD = 2147483135  # 2^31 - 1 - 512
 
-# fmt: off
 try:
     import requests
-    urllib3 = requests.packages.urllib3
-    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
     from requests.auth import AuthBase
     from requests.cookies import cookiejar_from_dict
 except ImportError:
     logger.error("Chosen backend requires 'requests' module\n")
     sys.exit(1)
 
-if sys.version_info[0] >= 3:
-    import io
-    def is_file(obj): return isinstance(obj, io.IOBase)
-    # prefer using monoatomic time if available
-    def get_time(): return time.monotonic()
-else:
-    def is_file(obj): return isinstance(obj, file)  # noqa pylint:disable=undefined-variable
-    def get_time(): return time.time()
-# fmt: on
-
-
-class AuthenticationError(Exception):
-    def __init__(self, msg):
-        super(AuthenticationError, self).__init__(msg)
-        self.msg = msg
-
-    def __str__(self):
-        return self.msg
-
-    def __repr__(self):
-        return self.__str__()
-
 
 class ProxmoxHTTPAuthBase(AuthBase):
+    def __call__(self, req):
+        return req
+
     def get_cookies(self):
         return cookiejar_from_dict({})
 
     def get_tokens(self):
         return None, None
 
+    def __init__(self, timeout=5, service="PVE", verify_ssl=False):
+        self.timeout = timeout
+        self.service = service
+        self.verify_ssl = verify_ssl
+
 
 class ProxmoxHTTPAuth(ProxmoxHTTPAuthBase):
     # number of seconds between renewing access tickets (must be less than 7200 to function correctly)
-    # if calls are made less frequently than 2 hrs, using the API token auth is reccomended
+    # if calls are made less frequently than 2 hrs, using the API token auth is recommended
     renew_age = 3600
 
-    def __init__(
-        self, base_url, username, password, otp=None, verify_ssl=False, timeout=5, service="PVE"
-    ):
+    def __init__(self, username, password, otp=None, base_url="", **kwargs):
+        super().__init__(**kwargs)
         self.base_url = base_url
         self.username = username
-        self.verify_ssl = verify_ssl
-        self.timeout = timeout
-        self.service = service
         self.pve_auth_ticket = ""
 
         self._get_new_tokens(password=password, otp=otp)
@@ -97,8 +79,12 @@ class ProxmoxHTTPAuth(ProxmoxHTTPAuthBase):
                     self.username, self.base_url + "/access/ticket"
                 )
             )
+        if response_data.get("NeedTFA") is not None:
+            raise AuthenticationError(
+                "Couldn't authenticate user: missing Two Factor Authentication (TFA)"
+            )
 
-        self.birth_time = get_time()
+        self.birth_time = time.monotonic()
         self.pve_auth_ticket = response_data["ticket"]
         self.csrf_prevention_token = response_data["CSRFPreventionToken"]
 
@@ -108,56 +94,38 @@ class ProxmoxHTTPAuth(ProxmoxHTTPAuthBase):
     def get_tokens(self):
         return self.pve_auth_ticket, self.csrf_prevention_token
 
-    def __call__(self, r):
+    def __call__(self, req):
         # refresh ticket if older than `renew_age`
-        if (get_time() - self.birth_time) >= self.renew_age:
-            logger.debug("refreshing ticket (age {0})".format(get_time() - self.birth_time))
+        time_diff = time.monotonic() - self.birth_time
+        if time_diff >= self.renew_age:
+            logger.debug(f"refreshing ticket (age {time_diff})")
             self._get_new_tokens()
 
         # only attach CSRF token if needed (reduce interception risk)
-        if r.method != "GET":
-            r.headers["CSRFPreventionToken"] = self.csrf_prevention_token
-        return r
-
-
-# DEPRECATED(1.1.0) - either use a password or the API Tokens
-class ProxmoxHTTPTicketAuth(ProxmoxHTTPAuth):
-    """Use existing ticket/token to create a session.
-
-    Overrides ProxmoxHTTPAuth so that an existing auth ticket and csrf token
-    may be used instead of passing username/password.
-    """
-
-    def __init__(self, auth_ticket, csrf_token):
-        self.pve_auth_ticket = auth_ticket
-        self.csrf_prevention_token = csrf_token
-        self.birth_time = get_time()
-
-        # deprecation notice
-        logger.warning(
-            "** Existing token auth is Deprecated as of 1.1.0\n** Please use the API Token Auth for long-running programs or pass existing ticket as password to the user/password auth"
-        )
+        if req.method != "GET":
+            req.headers["CSRFPreventionToken"] = self.csrf_prevention_token
+        return req
 
 
 class ProxmoxHTTPApiTokenAuth(ProxmoxHTTPAuthBase):
-    def __init__(self, username, token_name, token_value, service):
-        self.service = service
+    def __init__(self, username, token_name, token_value, **kwargs):
+        super().__init__(**kwargs)
         self.username = username
         self.token_name = token_name
         self.token_value = token_value
 
-    def __call__(self, r):
-        r.headers["Authorization"] = "{0}APIToken={1}!{2}{3}{4}".format(
+    def __call__(self, req):
+        req.headers["Authorization"] = "{0}APIToken={1}!{2}{3}{4}".format(
             self.service,
             self.username,
             self.token_name,
             SERVICES[self.service]["token_separator"],
             self.token_value,
         )
-        return r
+        return req
 
 
-class JsonSerializer(object):
+class JsonSerializer:
     content_types = [
         "application/json",
         "application/x-javascript",
@@ -182,6 +150,7 @@ class JsonSerializer(object):
             return {"errors": response.content}
 
 
+# pylint:disable=arguments-renamed
 class ProxmoxHttpSession(requests.Session):
     def request(
         self,
@@ -206,9 +175,12 @@ class ProxmoxHttpSession(requests.Session):
         a = auth or self.auth
         c = cookies or self.cookies
 
-        # take set verify flag from session if request does not have this parameter explicitly
+        # set verify flag from auth if request does not have this parameter explicitly
         if verify is None:
-            verify = self.verify
+            verify = a.verify_ssl
+
+        if timeout is None:
+            timeout = a.timeout
 
         # pull cookies from auth if not present
         if (not c) and a:
@@ -221,8 +193,11 @@ class ProxmoxHttpSession(requests.Session):
         for k, v in data.copy().items():
             # split qemu exec commands for proper parsing by PVE (issue#89)
             if k == "command":
-                data[k] = v if isinstance(v, list) else shell_split(v)
-            if is_file(v):
+                if isinstance(v, list):
+                    data[k] = v
+                elif "Windows" not in platform.platform():
+                    data[k] = shell_split(v)
+            if isinstance(v, io.IOBase):
                 total_file_size += get_file_size(v)
 
                 # add in filename from file pointer (patch for https://github.com/requests/toolbelt/pull/316)
@@ -235,26 +210,23 @@ class ProxmoxHttpSession(requests.Session):
                 # pylint:disable=import-outside-toplevel
                 from requests_toolbelt import MultipartEncoder
 
-                encoder = MultipartEncoder(fields=merge_dicts(data, files))
+                encoder = MultipartEncoder(fields={**data, **files})
                 data = encoder
                 files = None
                 headers = {"Content-Type": encoder.content_type}
             except ImportError:
                 # if the files will cause issues with the SSL 2GiB limit (https://bugs.python.org/issue42853#msg384566)
-                if total_file_size > 2147483135:  # 2^31 - 1 - 512
+                if total_file_size > SSL_OVERFLOW_THRESHOLD:
                     logger.warning(
                         "Install 'requests_toolbelt' to add support for files larger than 2GiB"
                     )
                     raise OverflowError("Unable to upload a payload larger than 2 GiB")
                 else:
                     logger.info(
-                        "Installing 'requests_toolbelt' will deacrease memory used during upload"
+                        "Installing 'requests_toolbelt' will decrease memory used during upload"
                     )
 
-        if not files and serializer:
-            headers = {"Content-Type": "application/x-www-form-urlencoded"}
-
-        return super(ProxmoxHttpSession, self).request(
+        return super().request(
             method,
             url,
             params,
@@ -273,7 +245,7 @@ class ProxmoxHttpSession(requests.Session):
         )
 
 
-class Backend(object):
+class Backend:
     def __init__(
         self,
         host,
@@ -284,10 +256,9 @@ class Backend(object):
         verify_ssl=True,
         mode="json",
         timeout=5,
-        auth_token=None,
-        csrf_token=None,
         token_name=None,
         token_value=None,
+        path_prefix=None,
         service="PVE",
     ):
 
@@ -297,7 +268,7 @@ class Backend(object):
                 if "]:" in host:
                     host, host_port = host.rsplit(":", 1)
             else:
-                host = "[{0}]".format(host)
+                host = f"[{host}]"
         elif ":" in host:
             host, host_port = host.split(":")
         port = host_port if host_port.isdigit() else port
@@ -306,30 +277,42 @@ class Backend(object):
         if not port:
             port = SERVICES[service]["default_port"]
 
-        self.base_url = "https://{0}:{1}/api2/{2}".format(host, port, mode)
+        self.mode = mode
+        if path_prefix is not None:
+            self.base_url = f"https://{host}:{port}/{path_prefix}/api2/{mode}"
+        else:
+            self.base_url = f"https://{host}:{port}/api2/{mode}"
 
-        if auth_token is not None:
-            # DEPRECATED(1.1.0) - either use a password or the API Tokens
-            self.auth = ProxmoxHTTPTicketAuth(auth_token, csrf_token)
-        elif token_name is not None:
+        if token_name is not None:
             if "token" not in SERVICES[service]["supported_https_auths"]:
                 config_failure("{} does not support API Token authentication", service)
 
-            self.auth = ProxmoxHTTPApiTokenAuth(user, token_name, token_value, service)
+            self.auth = ProxmoxHTTPApiTokenAuth(
+                user,
+                token_name,
+                token_value,
+                verify_ssl=verify_ssl,
+                timeout=timeout,
+                service=service,
+            )
         elif password is not None:
             if "password" not in SERVICES[service]["supported_https_auths"]:
                 config_failure("{} does not support password authentication", service)
 
             self.auth = ProxmoxHTTPAuth(
-                self.base_url, user, password, otp, verify_ssl, timeout, service
+                user,
+                password,
+                otp,
+                base_url=self.base_url,
+                verify_ssl=verify_ssl,
+                timeout=timeout,
+                service=service,
             )
-        self.verify_ssl = verify_ssl
-        self.mode = mode
-        self.timeout = timeout
+        else:
+            config_failure("No valid authentication credentials were supplied")
 
     def get_session(self):
         session = ProxmoxHttpSession()
-        session.verify = self.verify_ssl
         session.auth = self.auth
         # cookies are taken from the auth
         session.headers["Connection"] = "keep-alive"
@@ -391,14 +374,3 @@ def get_file_size_partial(file_obj):
     file_obj.seek(starting_cursor)
 
     return size
-
-
-def merge_dicts(*dicts):
-    """Compatibility polyfill for dict unpacking for python < 3.5
-    See PEP 448 for details on how merging functions
-
-    :return: merged dicts
-    :rtype: dict
-    """
-    # synonymous with {**dict for dict in dicts}
-    return {k: v for d in dicts for k, v in d.items()}
